

from pyexlatex.table.models.labels.label import Label

class CellSpacer(Label):

    def __init__(self, width=1):
        super().__init__(' ', span=width)

    def __repr__(self):
        return f'<CellSpacer(width={self.span})>'

    def __str__(self):
        if len(self) == 1:
            return str(self.value)
        else:
            return ' & '.join([self.value for i in range(len(self))])


import uuid
from typing import Tuple, Optional, Union, TYPE_CHECKING, Any, List

if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay

from pyexlatex.graphics.tikz.item import TikZItem
from pyexlatex.graphics.tikz.node.position.position import NodePosition
from pyexlatex.logic.format.contents import format_contents
from pyexlatex.graphics.tikz.node.position.directions import DirectionBase


class Node(TikZItem):
    """
    Represents a location in a graphic, but can also have a style and text, e.g. rectangle filled with text
    """
    label: Optional[str]

    def __init__(self, contents: Optional[Any] = None, location: Optional[Union[Tuple[int, int], DirectionBase, str]] = None,
                 label: Optional[str] = None,
                 options: Optional[List[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.add_data_from_content(location)
        self.location = NodePosition(location)
        self.add_data_from_content(contents)
        self.content = format_contents(contents)
        self.label = label if label is not None else str(uuid.uuid4())

        options = self._get_list_copy_from_list_or_none(options)
        if self.location.relative_location:
            options.append(str(self.location))

        super().__init__(self.get_name(), self.content, options=options, overlay=overlay)

    def get_name(self) -> str:
        if self.content is None:
            return 'coordinate'
        else:
            return 'node'

    @property
    def absolute_location_str(self) -> str:
        if not self.location.absolute_location:
            return ''

        return str(self.location)

    @property
    def content_str(self) -> str:
        text = self._empty_str_if_none(self._wrap_with_braces(self.content))
        return f'({self.label}) {self.absolute_location_str} {text}'

    def __str__(self):
        overlay = self.overlay if self.overlay is not None else ""
        item_str = fr'\{self.name}{overlay} {self.options_str} {self.content_str};'
        return item_str



from typing import Dict
from pyexlatex.models.references.bibtex.base import BibTexEntryBase

class BibTexEntry(BibTexEntryBase):

    def __init__(self, fields_dict: Dict[str, str]):
        self.optional_attrs = []
        self.fields = fields_dict
        super().__init__(self.item_accessor)

    @BibTexEntryBase.fields.setter  # type: ignore
    def fields(self, fields_dict: Dict[str, str]):
        rename_items = {
            'ENTRYTYPE': 'item_type',
            'ID': 'item_accessor'
        }
        skip_fields_items = ['item_type', 'item_accessor']
        for attr, value in fields_dict.items():
            if attr in rename_items:
                self_attr = rename_items[attr]
            else:
                self_attr = attr
            setattr(self, self_attr, value)
            # TODO [#16]: improve BibTexEntry fields property
            #
            # Come up with cleaner way to get added to fields output property
            if self_attr not in skip_fields_items:
                self.optional_attrs.append(self_attr)



from pyexlatex.models.item import SimpleItem


class Footnote(SimpleItem):
    """
    Creates a footnote at the bottom of the page, with a mark next to where the object itself is placed to
    reference the footnote.
    """
    name = 'footnote'

    def __init__(self, contents):
        super().__init__(self.name, contents)



from pyexlatex.models.item import NoOptionsNoContentsItem
from pyexlatex.models.documentitem import DocumentItem

class PrintBibliography(NoOptionsNoContentsItem, DocumentItem):
    name = 'printbibliography'

    def __init__(self):
        super().__init__(self.name)




from pyexlatex.models.section.base import TextAreaBase


class MultiColumn(TextAreaBase):
    """
    Layout multiple columns in a text area
    """
    name = 'multicols'

    def __init__(self, content, num_cols: int = 2, **kwargs):
        self.num_cols = num_cols
        self.add_package('multicol')
        super().__init__(self.name, content, env_modifiers=self._wrap_with_braces(str(self.num_cols)), **kwargs)



from typing import Union, List, Sequence

from pyexlatex.table.models.panels.collection import PanelCollection
from pyexlatex.table.models.table.row import Row
from pyexlatex.table.models.texgen.lines import TopRule, MidRule, BottomRule, TableLine
from pyexlatex.models.format.breaks import TableLineBreak, LineBreak

def build_tabular_content_from_panel_collection(panel_collection: PanelCollection, mid_rule=True):
    rows: List[Union[Row, TableLine]] = _build_tabular_rows_from_panel_collection(
        panel_collection=panel_collection,
        mid_rule=mid_rule
    )

    content: str = _build_tabular_str_from_rows_and_lines(rows)

    return content



def _build_tabular_rows_from_panel_collection(panel_collection: PanelCollection, mid_rule=True):
    rows: List[Union[Row, TableLine]] = [TopRule()]
    panels = list(panel_collection.iterpanels())
    for i, panel in enumerate(panels):
        rows += panel.rows
        # add mid rule when:
        # boolean is turned on,
        # not the last loop,
        # and panel is not made entirely of spacers,
        # and next panel is not made entirely of spacers
        if mid_rule and i != len(panels) - 1 and not panel.is_spacer and not panels[i + 1].is_spacer:
            rows.append(MidRule())

    rows.append(BottomRule())
    return rows


def _build_tabular_str_from_rows_and_lines(rows_and_lines: Sequence[Union[Row, TableLine]], break_size_adjustment: str=None):
    output_str = ''
    for i, row_or_line in enumerate(rows_and_lines):
        end = _get_break_by_type_of_instance(row_or_line, break_size_adjustment=break_size_adjustment)
        output_str += f'{row_or_line}{end}'
    return output_str


def _get_break_by_type_of_instance(row_or_line: Union[Row, TableLine], break_size_adjustment: str=None):
    table_row_break = TableLineBreak(break_size_adjustment)
    line_break = LineBreak()
    if isinstance(row_or_line, TableLine):
        end = line_break
    elif isinstance(row_or_line, Row):
        end = table_row_break
    else:
        raise NotImplementedError(f'could not determine type of break for row or line type {type(row_or_line)}')

    return end




from pyexlatex.models.item import EqualsItem


class HangIndent(EqualsItem):
    name = 'hangindent'

    def __init__(self, hang_by: float = 1):
        self.hang_by = hang_by
        super().__init__(self.name, f'{self.hang_by}cm')



from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.item import SimpleItem


class Italics(TextAreaMixin, SimpleItem):
    """
    Italicizes text.
    """
    name = 'textit'

    def __init__(self, contents, **kwargs):
        super().__init__(self.name, contents, **kwargs)



from pyexlatex.models.item import SimpleItem, NoOptionsNoContentsItem
from pyexlatex.models.page.footer import RightFooter, CenterFooter
from pyexlatex.logic.builder import _build


class PageReference(SimpleItem):
    name = 'pageref'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class ThisPageNumber(NoOptionsNoContentsItem):
    """
    Gets a reference to the current page number, can be included in headers and more.
    """
    name = 'thepage'

    def __init__(self):
        super().__init__(self.name)


last_page = PageReference('LastPage')
this_page = r'\thepage\ '  # TODO: use ThisPageNumber
page_reference_str = f'Page {this_page} of {last_page}'
right_aligned_page_numbers = _build([
    RightFooter(page_reference_str),  # add page number to bottom right
    CenterFooter('')  # cancel original page number at bottom center
])




from pyexlatex.models.item import SimpleItem

class ExtraFloats(SimpleItem):
    name = 'extrafloats'

    def __init__(self, num_extra: int = 1000):
        super().__init__(self.name, str(num_extra))



from typing import Sequence
from pyexlatex.models.item import MultiOptionSimpleItem


class SetBeamerColor(MultiOptionSimpleItem):
    name = 'setbeamercolor'

    def __init__(self, beamer_item: str, color_options: Sequence[str], **kwargs):
        color_options_str = ', '.join(color_options)
        super().__init__(self.name, beamer_item, color_options_str, **kwargs)



from pyexlatex.models.mixins import StringAdditionMixin
from pyexlatex.models.format.breaks import LineBreak

# TODO [#20]: Make new column types more flexible.
#
# Currently just passes full strings

class NewColumnTypes(StringAdditionMixin):

    def __init__(self):
        self.contents: [NewColumnType] = [
            NewColumnType(r'\newcolumntype{L}[1]{>{\raggedright\let\newline\\\arraybackslash\hspace{0pt}}m{#1}}'),
            NewColumnType(r'\newcolumntype{C}[1]{>{\centering\let\newline\\\arraybackslash\hspace{0pt}}m{#1}}'),
            NewColumnType(r'\newcolumntype{R}[1]{>{\raggedleft\let\newline\\\arraybackslash\hspace{0pt}}m{#1}}'),
            NewColumnType(r'\newcolumntype{.}{D{.}{.}{-1}}')
        ]

    def __str__(self):
        return LineBreak().join(self.contents)



class NewColumnType(str):
    pass


from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.models.commands.begin import Begin
from pyexlatex.models.commands.end import End


class Environment:

    def __init__(self, name, modifiers: Optional[str] = None, overlay: Optional['Overlay'] = None):
        self.name = name

        if modifiers is None:
            modifiers = ''
        if overlay is not None:
            modifiers = str(overlay) + modifiers

        self._begin = Begin(name, modifiers=modifiers)
        self._end = End(name)

    def __repr__(self):
        return f'<Environment(name={self.name})>'

    def wrap(self, other):
        from pyexlatex.logic.builder import _build
        return _build([self._begin, other, self._end])



from copy import deepcopy

from pyexlatex.table.models.spacing.cell import CellSpacer
from pyexlatex.table.models.table.section import TableSection
from pyexlatex.table.models.data.valuestable import ValuesTable

class ColumnPadTable(ValuesTable):

    def __init__(self, width: int=1):
        self.width = width
        super().__init__([])

    @property
    def num_columns(self):
        return 1

    def __repr__(self):
        return f'<ColumnPadTable({len(self.rows)})>'

    def __add__(self, other):

        if not isinstance(other, TableSection):
            raise NotImplementedError

        out_section: TableSection = deepcopy(other)

        for row in out_section.rows:
            row.pad(other.num_columns + self.width, direction='left')

        return out_section






from pyexlatex.models.mixins import StringAdditionMixin
from mixins.repr import ReprMixin


class Break(StringAdditionMixin):
    pass

class LineBreak(Break):

    def __init__(self):
        pass

    def __str__(self):
        return '\n'


class TableLineBreak(Break, ReprMixin):
    """
    A line break character to be used in LaTeX tables
    """
    repr_cols = ['size_adjustment']

    def __init__(self, size_adjustment: str=None):
        self.size_adjustment = size_adjustment

    def __str__(self):
        return _table_break_str(self.size_adjustment, add_line_break=True)


class OutputLineBreak(Break, ReprMixin):
    """
    Create an intentional line break in text
    """
    repr_cols = ['size_adjustment']

    def __init__(self, size_adjustment: str=None):
        self.size_adjustment = size_adjustment

    def __str__(self):
        return _table_break_str(self.size_adjustment)


def _table_break_str(size_adjustment: str=None, add_line_break: bool = False):
    if size_adjustment:
        size_adjustment_str = f'[{size_adjustment}]'
    else:
        size_adjustment_str = ''
    output_str = r'\\' + size_adjustment_str
    if add_line_break:
        output_str += LineBreak()
    return  output_str


from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.documentitem import DocumentItem


class DataStore(ContainerItem, DocumentItem):
    """
    Holds LaTeX item data and can be included directly in documents. Mainly for internal use.
    """

    def __init__(self, content=None):
        if content is not None:
            self.add_data_from_content(content)
        else:
            self.init_data()

    def __str__(self):
        return ''



from mixins.repr import ReprMixin
from pyexlatex.table.models.mixins.addvalues.row import RowAddMixin
from pyexlatex.texgen.replacements.file import general_latex_replacements


class DataItem(ReprMixin, RowAddMixin):
    repr_cols = ['value']

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f'{general_latex_replacements(str(self.value))}'

    def _add_class(self, other):
        from pyexlatex.table.models.table.row import Row
        from pyexlatex.table.models.data.row import DataRow
        # keep same class if both are same class
        # otherwise, default to Row class
        self_class = type(self)
        other_class = type(other)
        klass = DataRow if self_class == other_class else Row

        return klass

    def __len__(self):
        return 1


from typing import Union, Sequence
from pyexlatex.models.template import Template
from pyexlatex.graphics.tikz.node.node import Node
from pyexlatex.graphics.tikz.scope import Scope
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.graphics.tikz.picture import TikZPicture
from pyexlatex.graphics.shape import Shape
from pyexlatex.graphics.tikz.item import TikZItem


class ModifiedPicture(Template):

    def __init__(self, picture: Union[str, Graphic, TikZPicture], draw_items: Sequence[Union[TikZItem, Shape]],
                 add_grid: bool = False):
        self.picture = picture
        self.draw_items = draw_items
        self.add_grid = add_grid
        self.contents = self._get_contents()
        super().__init__()

    def _get_graphic(self) -> Union[Graphic, TikZPicture]:
        if isinstance(self.picture, (TikZPicture, Graphic)):
            return self.picture

        # Assuming str of filepath was passed
        return Graphic(self.picture)

    def _get_contents(self):
        # Add the image as the contents of a node
        image_node = Node(self._get_graphic(), options=['anchor=south west'])

        if self.add_grid:
            # TODO [#4]: use models in drawing help lines grid
            draw_items = [
                r'\draw[help lines,xstep=.1,ystep=.1] (0,0) grid (1,1);',
                r'\foreach \x in {0,1,...,9} { \node [anchor=north] at (\x/10,0) {0.\x}; }',
                r'\foreach \y in {0,1,...,9} { \node [anchor=east] at (0,\y/10) {0.\y}; }',
                *self.draw_items
            ]
        else:
            draw_items = self.draw_items

        # Create a scope where the bounding box is the image node, now with a relative coordinate system (0, 1)
        scope = Scope(
            draw_items,
            options=[
                f'x={self._wrap_with_braces(f"({image_node.label}.south east)")}',
                f'y={self._wrap_with_braces(f"({image_node.label}.north west)")}'
            ]
        )

        picture = TikZPicture([
            image_node,
            scope
        ])

        return picture




from typing import Optional
from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.presentation.beamer.theme.setcolor import SetBeamerColor


class BlockBase(TextAreaBase):
    name = '<invalid, use a subclass, not BlockBase>'

    def __init__(self, content, title: Optional[str] = None, text_color: Optional[str] = None,
                 header_color: Optional[str] = None, **kwargs):
        self.content = content
        self.title = title
        title_str = title if title is not None else ''
        title_modifier_str = f'{{{title_str}}}'

        pre_env_contents = None
        post_env_contents = None
        if text_color is not None or header_color is not None:
            from pyexlatex.logic.builder import _build
            color_options = []
            if text_color is not None:
                color_options.append(f'fg={text_color}')
            if header_color is not None:
                color_options.append(f'bg={header_color}')
            set_color = SetBeamerColor('block title', color_options)
            pre_env_contents = _build([
                '{',
                set_color
            ])
            post_env_contents = '}'


        super().__init__(self.name, self.content, env_modifiers=title_modifier_str,
                         pre_env_contents=pre_env_contents,
                         post_env_contents=post_env_contents, **kwargs)


class Block(BlockBase):
    """
    Block with an optional header
    """
    name = 'block'


class AlertBlock(BlockBase):
    """
    Block with an optional header which is red
    """
    name = 'alertblock'


class ExamplesBlock(BlockBase):
    """
    Block which always has the header title "Examples"
    """
    name = 'examples'

    def __init__(self, content):
        super().__init__(content, None)  # examples does not support a title



from typing import Any

from pyexlatex import Center
from pyexlatex.models.format.adjustbox import AdjustBox
from pyexlatex.models.sizes.textheight import TextHeight
from pyexlatex.models.sizes.textwidth import TextWidth


def adjust_to_full_size_and_center(content: Any) -> Center:
    """
    Takes content and adjusts it to 90% text width, 80% text height, but keeping the aspect ratio.
    """

    content = adjust_to_size(content, 0.9, 0.8, keep_aspect_ratio=True)
    content = Center(content)
    return content


def adjust_to_size(content: Any, width: float, height: float, keep_aspect_ratio: bool = True
                   ) -> AdjustBox:
    adjust_options = [
        f'width={width}{TextWidth()}',
        f'height={height}{TextHeight()}',
    ]
    if keep_aspect_ratio:
        adjust_options.append('keepaspectratio')

    return AdjustBox(
        content,
        adjust_options=adjust_options
    )


from pyexlatex.models.template import Template
from pyexlatex.models.format.paragraph.raggedright import RaggedRight
from pyexlatex.models.format.paragraph.sloppy import Sloppy
from pyexlatex.models.format.text.nohyphens import NoHyphens


class NoLineBreak(Template):
    """
    Prevents a line break within text, where a hyphen would normally be placed.

    Notes:
        Using this will have side effects to the entire paragraph, depending on the alternative chosen. See
        the docstring for alternative.
    """
    name = 'nohyphens'

    def __init__(self, contents, alternative: str = 'ragged'):
        """

        :param contents:
        :param alternative: 'ragged' or 'sloppy'. As the word won't be broken, there must be some alternative.
            Passing 'ragged' will make all the other words on the first line stay in the same position, so that
            a gap will appear at the end of the line where the NoLineBreak word would have started originally. Passing
            'sloppy' will make it so that the first line's words spacing gets stretched so that there is still right
            alignment on that line, despite the NoLineBreak word being moved to the next line. Note: This will affect
            the entire paragraph, not just the NoLineBreak word.
        """
        self.alternative = alternative.lower()
        self.orig_contents = contents
        self._validate()

        self.contents = self._get_contents()

        super().__init__()

    def _get_contents(self):
        return [
            self._alternative_obj,
            NoHyphens(self.orig_contents)
        ]

    @property
    def _alternative_obj(self):
        if self.alternative == 'ragged':
            return RaggedRight()
        elif self.alternative == 'sloppy':
            return Sloppy()

    def _validate(self):
        self._validate_alternative()

    def _validate_alternative(self):
        alternatives = (
            'ragged',
            'sloppy',
        )
        if self.alternative not in alternatives:
            raise ValueError(f'Must pass valid alternative, one of {alternatives} to NoLineBreak. Got {self.alternative}')




from typing import Optional
from pyexlatex.models.sizes.textwidth import TextWidth
from pyexlatex.models.sizes.textheight import TextHeight
from pyexlatex.models.boxes.vbox import VBox
from pyexlatex.models.section.base import TextAreaBase


class Column(TextAreaBase):
    name = 'column'
    repr_cols = [
        'contents'
        'frac_of_text_width',
        'frac_of_page_height'
    ]

    def __init__(self, content, frac_of_text_width: float = 1.0, frac_of_page_height: Optional[float] = 0.8):
        self.frac_of_text_width = frac_of_text_width
        self.add_data_from_content(content)

        if frac_of_page_height is not None:
            contents = VBox(content, size_str=f' to {frac_of_page_height}{TextHeight()}')
        else:
            contents = content

        super().__init__(self.name, contents, env_modifiers=self.text_width_str)

    @property
    def text_width_str(self) -> str:
        return self._wrap_with_braces(f'{self.frac_of_text_width}{TextWidth()}')


from typing import Sequence
from pyexlatex.models.lists.base import ListBase
from pyexlatex.models.control.setcounter import SetCounter


class OrderedList(ListBase):
    """
    Numbered, ordered items.
    """
    name = 'enumerate'
    is_ListBase = True  # so that passing within lists will treat as a list

    def __init__(self, items: Sequence, initial_number: int = 1, **kwargs):
        self.initial_number = initial_number

        if self.initial_number != 1:
            items = [SetCounter('enumi', self.initial_number - 1)] + list(items)

        super().__init__(items, **kwargs)




from typing import Optional, Type
from pyexlatex.models.item import Item
from mixins.repr import ReprMixin
from pyexlatex.logic.format.contents import format_contents as fmt
from pyexlatex.logic.builder import _build
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.models.label import Label
from pyexlatex.models.containeritem import ContainerItem


class TextAreaMixin(ContainerItem):
    """
    Mixin for extracting data from content then formatting it, regardless of the data type passed
    """
    name = 'textarea'
    next_level_down_class: Optional[Type['TextAreaBase']] = None  # once subclassed, will be overridden with the next level down text area class

    def __init__(self, name, contents, label: Optional[str] = None, **kwargs):
        self.add_data_from_content(contents)
        contents = self.format_contents(contents)
        if label is not None:
            label = Label(label)
            if not isinstance(contents, (list, tuple)):
                contents = [contents]
            contents = contents + [label]
        self.contents = contents
        super().__init__(name, contents, **kwargs)

    def format_contents(self, contents):
        if isinstance(contents, (list, tuple)):
            return [self.format_contents(c) for c in contents]
        elif isinstance(contents, dict):
            if self.next_level_down_class is None:
                raise ValueError(f'cannot parse dict as have no next_level_down_class in {self.__class__.name}')
            subcontents = []
            for title, content in contents.items():
                subcontents.append(
                    self.next_level_down_class(content, title=title)
                )
            return subcontents
        else:
            # Not an iterable
            return self._format_content(contents)


    def _format_content(self, content):
        if isinstance(content, str):
            return fmt(content)
        else:
            # Class is responsible for formatting. This may be a latex class or some
            # other harmless conversion such as int. It may also be an issue if the __str__
            # method of the class is not valid latex
            return content


class TextAreaBase(TextAreaMixin, Item, ReprMixin):
    repr_cols = ['title', 'contents']


class SectionBase(TextAreaBase):
    name = 'section'
    repr_cols = ['title', 'short_title', 'contents']

    def __init__(self, contents, title: str, short_title: Optional[str] = None, **kwargs):
        self.title = title
        self.short_title = short_title
        super().__init__(self.name, contents, env_modifiers=self.env_modifiers, **kwargs)

    @property
    def env_modifiers(self):
        modifier_str = ''
        if self.short_title is not None:
            modifier_str += f'[{fmt(self.short_title)}]'

        modifier_str += f'{{{fmt(self.title)}}}'

        return modifier_str


class ParagraphBase(TextAreaBase):
    name = 'paragraph'
    repr_cols = ['title', 'contents']

    def __init__(self, contents, title: Optional[str] = None, **kwargs):
        self.title = title
        super().__init__(self.name, contents, env_modifiers=self.env_modifiers, **kwargs)

    @property
    def env_modifiers(self):
        if self.title is not None:
            return f'{{{self.title}}}'

        return None


from pyexlatex.models.control.documentclass.classtypes.custom.resume import resume_class_type

CUSTOM_CLASS_TYPE_OBJS = [
    resume_class_type
]

CUSTOM_CLASS_TYPES = {class_type.name: class_type for class_type in CUSTOM_CLASS_TYPE_OBJS}



from pyexlatex.models.item import MultiOptionSimpleItem
from mixins.attrequals import EqOnAttrsWithConversionMixin, EqHashMixin


class NewEnvironment(MultiOptionSimpleItem, EqOnAttrsWithConversionMixin, EqHashMixin):
    name = 'newenvironment'
    equal_attrs = ['env_name', 'begin', 'end']
    convert_type = str

    def __init__(self, env_name: str, begin_env: str, end_env: str):
        self.env_name = env_name
        self.begin = begin_env
        self.end = end_env
        super().__init__(self.name, env_name, begin_env, end_env)




from typing import Union, Optional
from sympy import Eq

from pyexlatex.models.item import IsLatexItemMixin
from pyexlatex.models.mixins import IsSpecificClassMixin
from pyexlatex.logic.format import eq as format_eq

class Equation(IsSpecificClassMixin, IsLatexItemMixin):
    """
    Pass sympy or string equations to have them rendered in LaTeX.
    """
    name = 'equation'

    def __init__(self, eq: Optional[Eq] = None, str_eq: Optional[str] = None,
                 inline: bool = True):
        self._validate(eq, str_eq)
        self.eq_str = eq if eq else str_eq
        self.inline = inline
        super().__init__()

    def __str__(self):
        return self.formatted_eq

    @property
    def formatted_eq(self):
        if self.inline:
            return format_eq.inline(self.eq_str)
        else:
            return format_eq.offset(self.eq_str)

    def _validate(self, eq: Optional[Eq] = None, str_eq: Optional[str] = None):
        if eq is None and str_eq is None:
            raise ValueError('must pass one of eq or str_eq')
        if eq is not None and str_eq is not None:
            raise ValueError('must pass at most one of eq or str_eq')


    @property
    def eq_str(self):
        return self._eq_str

    @eq_str.setter
    def eq_str(self, value: Union[Eq, str]):
        if isinstance(value, str):
            # already formatted
            self._eq_str = value
        else:
            from sympy import latex
            self._eq_str = latex(value)



from typing import Optional
from pyexlatex.models.template import Template
from pyexlatex.models.format.text.bold import Bold
from pyexlatex.models.format.breaks import OutputLineBreak
from pyexlatex.layouts.spaced import HorizontallySpaced
from pyexlatex.models.format.nopagebreak import NoPageBreak


class Award(Template):
    """
    Represents an award or grant in a resume.
    """
    SPACE_BETWEEN_ADJUSTMENT = '-8pt'

    def __init__(self, award_name: str, received: Optional[str] = None, extra_info: Optional[str] = None,
                 prevent_page_break: bool = True):
        self.award_name = award_name
        self.received = received
        self.extra_info = extra_info
        self.prevent_page_break = prevent_page_break
        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        items = [item for item in [self._award_obj, self.received] if item is not None]
        items = [
            HorizontallySpaced(items),
            OutputLineBreak(size_adjustment=self.SPACE_BETWEEN_ADJUSTMENT)
        ]
        if self.prevent_page_break:
            items = NoPageBreak(items)
        return items

    @property
    def _award_obj(self):
        if self.extra_info is None:
            return self._award_formatter(self.award_name)

        return [self._award_formatter(self.award_name), f' ({self.extra_info})']

    def _award_formatter(self, award_str: str):
        return Bold(award_str)



from pyexlatex.figure.packages import default_packages as default_figure_packages
from pyexlatex.table.models.texgen.packages import default_packages as default_table_packages

# exclude geometry package as has different margins in figure vs table. Use table default.
figure_packages = [package for package in default_figure_packages if not package.matches_name('geometry')]

# Remove duplicates, keeping order
default_packages = list({package: '' for package in default_table_packages + figure_packages}.keys())



from typing import Union, Iterable, List
import re
from copy import deepcopy

from pyexlatex.logic.tools import _add_if_not_none
from pyexlatex.table.models.labels.label import Label
from pyexlatex.table.models.mixins.addvalues.row import RowAddMixin
from pyexlatex.table.models.table.rowbase import RowBase


class LabelCollection(RowBase):
    """
    Represents one row of labels. Use to construct a LabelTable to apply to a DataTable.

    Main usage is LabelCollection.from_str_list
    """
    repr_cols = ['values', 'underlines']

    def __init__(self, values: List[Label], underline: Union[int, str]=None):
        """

        :param values:
        :param underline: int, str, or None, pass index or label to underline, pass a str range, e.g. '2-5' for a
                          range of labels to underline, pass space separated indices as a str, e.g. '0 2' for
                          separated underlines, or a combination, e.g. '2-5 8'
        """
        if isinstance(values, tuple):
            values = list(values)
        self.values = values
        underline_label_indices = _convert_underline_to_label_index_list(underline)
        self.underlines = self._convert_label_indices_to_column_indices(underline_label_indices)

    def __str__(self):
        return str(sum(self.values))

    def matches(self, other):
        """
        Compare on the basis of having same values, rather than same instance
        Use regular equality to test if same instance
        :param other:
        :return:
        """
        for i, value in enumerate(self):
            try:
                other[i]
            except IndexError:
                return False # any one misalignment, no match
            if value != other[i]:
                return False

        # same number of rows, all rows equal
        return True

    @classmethod
    def from_str_list(cls, str_list: List[str], underline: Union[int, str]=None) -> 'LabelCollection':
        """

        Args:
            str_list:
            underline: int, str, or None, pass index or label to underline, pass a str range, e.g. '2-5' for a
                  range of labels to underline, pass space separated indices as a str, e.g. '0 2' for
                  separated underlines, or a combination, e.g. '2-5 8'

        Returns:

        """
        labels = [Label(value) for value in str_list]
        return cls(labels, underline=underline)

    @classmethod
    def parse_unknown_type(cls, unknown_type: Union[str, Iterable[str], 'LabelCollection']) -> 'LabelCollection':
        if isinstance(unknown_type, LabelCollection):
            return unknown_type
        if isinstance(unknown_type, str):
            unknown_type = [unknown_type]
        if isinstance(unknown_type, list):
            return LabelCollection.from_str_list(unknown_type)
        else:
            raise ValueError(f'unable to parse type {type(unknown_type)} into label collection')

    def _convert_label_indices_to_column_indices(self, label_indices: List[int]):
        column_indices: List[int] = []
        position = 0
        for i, value in enumerate(self.values):
            begin_position = position
            position += _get_item_length(value)
            # if this label is included in label indices, add all ints in the range of span of label
            if i in label_indices:
                column_indices += [i for i in range(begin_position, position)]

        if column_indices == []:
            return None

        return column_indices

    def shift_underlines(self, shift):
        if self.underlines is None:
            return
        self.underlines = [u + shift for u in self.underlines]

    def __add__(self, other):
        # need to handle shifting of indices
        if isinstance(other, LabelCollection):
            to_add = deepcopy(other)
            to_add.shift_underlines(len(self))
        else:
            to_add = other

        result = super().__add__(to_add)

        # carry through underlines with addition
        if isinstance(result, LabelCollection):
            underlines = _add_if_not_none(self.underlines, to_add.underlines)
            result.underlines = underlines

        return result

    def __radd__(self, other):
        # need to handle shifting of indices
        to_add = deepcopy(self)
        to_add.shift_underlines(len(self))

        result = RowAddMixin.radd(to_add, other)

        # carry through underlines with addition
        if isinstance(result, LabelCollection):
            if isinstance(other, LabelCollection):
                underlines = _add_if_not_none(to_add.underlines, other.underlines)
            else:
                underlines = to_add.underlines
            result.underlines = underlines

        return result

    def pad(self, length: int, direction='right'):
        """
        Expand row out to the right or left with blanks, until it is length passed
        :param length:
        :return:
        """

        # only necessary to move underline columns if padding left
        if direction == 'left':
            num_values_to_add = length - len(self)
            self.shift_underlines(num_values_to_add)

        super().pad(length=length, direction=direction)

def _get_item_length(item):
    if isinstance(item, Label):
        return len(item)
    else:
        return 1

def _convert_underline_to_label_index_list(underline: Union[int, str]=None):
    if underline is None:
        return []

    assert isinstance(underline, (int, str))

    if isinstance(underline, int):
        return [underline]

    # underline is str
    int_list: List[int] = []
    for part in underline.split():
        # handle each space separated part. if just a range, will only be one part
        # check if is range like '3-5'
        try:
            if _is_range_str(part):
                int_list += _range_str_to_int_list(part)
            else:
                int_list.append(int(part))
        except ValueError:
            raise NotImplementedError(f'could not parse underline str into int. full underline '
                                      f'str: {underline}. failed processing part: {part}')

    return int_list

def _range_str_to_int_list(underline: str):
    bottom, top = underline.split('-')
    return [i for i in range(int(bottom), int(top) + 1)]


def _is_range_str(underline: str):
    pattern = re.compile(r'\d+-\d+')
    if pattern.match(underline):
        return True
    else:
        return False



from typing import Optional, Any
from pyexlatex.models.item import Item
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.lists.item import ListItem
from pyexlatex.models.format.nopagebreak import NoPageBreak


class Employment(ContainerItem, Item):
    name = 'employment'

    def __init__(self, contents, company_name: str, employed_dates: str, job_title: str, location: str,
                 extra_contents: Optional[Any] = None, prevent_page_break: bool = True):
        self.company_name = company_name
        self.employed_dates = employed_dates
        self.job_title = job_title
        self.location = location
        self.extra_contents = extra_contents
        self.prevent_page_break = prevent_page_break

        self.add_data_from_content([
            contents,
            company_name,
            employed_dates,
            job_title,
            location,
            extra_contents
        ])

        if not isinstance(contents, (list, tuple)):
            contents = [contents]

        contents = [ListItem(content) for content in contents]

        if extra_contents is not None:
            contents.append(extra_contents)

        if self.prevent_page_break:
            contents = NoPageBreak(contents)

        super().__init__(
            self.name,
            contents,
            env_modifiers=self._modifier_str
        )

    @property
    def _modifier_str(self) -> str:
        output = ''
        for item in [self.company_name, self.employed_dates, self.job_title, self.location]:
            output += self._wrap_with_braces(item)
        return output



from pyexlatex.models.item import MultiOptionSimpleItem


class SetCounter(MultiOptionSimpleItem):
    name = 'setcounter'

    def __init__(self, set_for: str, value: int):
        self.set_for = set_for
        self.value = value

        super().__init__(self.name, set_for, str(value))



import posixpath

from pyexlatex.figure.models import Subfigure, Figure
from pyexlatex.logic.pdf.main import document_to_pdf_and_move
from pyexlatex.texgen.replacements.filename import _latex_valid_basename
from pyexlatex.models.document import Document
from pyexlatex.models.package import Package


def filepaths_to_pdf_figure_and_move(filepaths,  outfolder, outname='figure', as_document=True,
                                     subfigure_kwargs={}, figure_kwargs={}, document_kwargs={}):

    basenames = [_latex_valid_basename(filepath) for filepath in filepaths]
    sources_paths = [posixpath.join('Sources', basename) for basename in basenames]

    document_or_figure = _filepaths_to_document_or_figure(
        sources_paths,
        subfigure_kwargs=subfigure_kwargs,
        figure_kwargs=figure_kwargs,
        document_kwargs=document_kwargs,
        as_document=as_document
    )

    document_to_pdf_and_move(
        document_or_figure,
        outfolder,
        image_paths=filepaths,
        outname=outname,
        as_document=as_document
    )

    return document_or_figure


def _filepaths_to_document_or_figure(filepaths, subfigure_kwargs={}, figure_kwargs={}, document_kwargs={},
                                     as_document=True):
    subfigures = [Subfigure(fp, **subfigure_kwargs) for fp in filepaths]
    figure = Figure(subfigures, **figure_kwargs)

    if not as_document:
        return figure

    simple_package_strs = [
        'caption',
        'subcaption',
        'graphicx',
        'pdflscape'
    ]
    simple_packages = [Package(str_) for str_ in simple_package_strs]

    packages = simple_packages + []  # add any packages with options here with the Package class

    document = Document(
        figure,
        packages,
        **document_kwargs
    )

    return document




import datetime
import filecmp
import os

from pyexlatex.tools.fileops import move_if_exists, remove_all_if_exist, remove_if_exists, move_all_if_exists


def date_time_move_latex(tablename,filepath, folder_name='Tables'):
    r"""
    Takes a LaTeX tex and PDF after the PDF's been created by pdflatex and moves it into a table
    folder with the date, checking if the table was just previously created. If it's the same as
    before, it just deletes the file.

    Required arguments:
        tablename: operating system name of the table, without extensions
        filepath: full filepath of table, without table name. put r before quotes as follows: r'C:\Users\Folder'

    """

    def exit_sequence():
        inpath_aux = os.path.join(filepath, str(tablename) + '.aux')
        inpath_log = os.path.join(filepath, str(tablename) + '.log')
        remove_all_if_exist([inpath_aux, inpath_log])
        return

    def join_paths(*paths):
        return os.path.abspath(os.path.join(*paths))

    table_pdf = tablename + ".pdf"
    table_tex = tablename + ".tex"
    table_xlsx = tablename + ".xlsx"
    inpath_pdf = join_paths(filepath,table_pdf)
    inpath_tex = join_paths(filepath,table_tex)
    inpath_xlsx = join_paths(filepath,table_xlsx)
    all_inpaths = [inpath_pdf, inpath_tex, inpath_xlsx]

    tables_path = join_paths(filepath, folder_name) #set table directory
    if not os.path.exists(tables_path): #create a general table directory if it doesn't exist
        os.makedirs(tables_path)

    current_time = datetime.datetime.today().timetuple()
    format_time = [str(current_time[0]),str(current_time[1]),str(current_time[2])]
    for i in range(3):
        if current_time[i] < 10:
            format_time[i] = "0" + str(current_time[i])
    datetime_str = "{}-{}-{}_".format(format_time[0],format_time[1],format_time[2])

    count = 0 #set up count variable
    while True: #continuous loop
        count += 1
        str_count = "Num" + str(count)
        name_str = datetime_str + str_count
        folder_path = os.path.join(tables_path,name_str)
        outpath_tex =  os.path.join(folder_path, table_tex)
        if os.path.exists(folder_path): #if the folder already exists
            if os.path.exists(outpath_tex): #if there is already a tex file with the same name
                if filecmp.cmp(outpath_tex,inpath_tex) == True: #if this is the same exact table
                    exit_sequence()
                    remove_all_if_exist(all_inpaths)
                    return None
                else: #if there is a tex file with the same name but it's not the same table
                    continue #go to next iteration of loop (change output number)
            else:
                move_all_if_exists(all_inpaths, folder_path)
                exit_sequence()
                return folder_path
        else: #if the folder doesn't exist
            os.mkdir(folder_path) #create the folder
            move_all_if_exists(all_inpaths, folder_path)
            exit_sequence()
            return folder_path


def csv_to_raw_latex(infile, csvstring=False, missing_rep=" - ", formatstr='{:.3f}', skipfix=None):
    '''
    Takes a CSV text file and converts it to a LaTeX formatted list, with each line of the LaTeX
    file as an item in the list.

    Required options:
        infile: Full file path of CSV (include r before quotes)

    Optional options:
        csvstring: True to pass a CSV string to infile rather than load from file
        missing_rep: Representation for missing numbers, default " - "
        formatstr: Python string for number formatting, for example '{:.3f}' with quotes
        skipfix: String or list of strings of fixes to skip, options are ['&','%','_']
    '''
    latex_list = []
    if not csvstring:
        f = open(infile,'r')
    else:
        from io import StringIO
        f = StringIO(infile)

    if skipfix:
        if isinstance(skipfix, str):
            skipfix = [skipfix]
        assert isinstance(skipfix, list)

    csv_list = f.readlines()
    miss_csv_list = replace_missing_csv(csv_list,missing_rep)
    latex_list = []
    for i, line in enumerate(miss_csv_list):
        line_string = ''
        for j, item in enumerate(line):
            if j != 0: #don't put an & before the first item in line
                line_string += ' & '
            #LaTeX character fixes
            if skipfix:
                if '&' not in skipfix:
                    item = item.replace('&', '\&')
                if '%' not in skipfix:
                    item = item.replace('%', '\%')
                if '_' not in skipfix:
                    item = item.replace('_', '\_')
            else: #make all replacements
                item = item.replace('&','\&')
                item = item.replace('%','\%')
                item = item.replace('_','\_')
            if item.find('.') is not -1: #if we are dealing with a number with decimals
                if formatstr:
                    try:
                        item = formatstr.format(float(item))
                    except:
                        pass
            item = item.replace('\n','')
            line_string += item
        line_string += " \\\ \n"
        if i == 0: #on the first line, remove quotes from names
            line_string = line_string.replace('''"''','') #strip out quotes
        latex_list.append(line_string)

    if not csvstring:
        f.close()

    return latex_list


def replace_missing_csv(csv_list, missing_rep):
    '''
    Replaces missing items in a CSV with a given missing representation string.
    '''
    full_list = []
    for line in csv_list:
        line_list = line.split(',')
        new_line_list = []
        for item in line_list:
            if item == '': #if the item is missing
                item = missing_rep
            new_line_list.append(item)
        full_list.append(new_line_list)
    return full_list


from typing import Optional, Callable, Tuple
import warnings
from pyexlatex.logic.pdf.errors.exc import TooManyUnprocessedFloatsException
from pyexlatex.models.control.extrafloats import ExtraFloats

EXTRA_FLOATS_LIMIT = 10000
DEFAULT_EXTRA_FLOATS = 1000
FLOATS_INCREASE_FACTOR = 2

def get_extra_float_and_new_num_floats(extra_floats_num: Optional[int] = None) -> Tuple[ExtraFloats, int]:
    if extra_floats_num is None:
        num_floats = DEFAULT_EXTRA_FLOATS
    else:
        num_floats = extra_floats_num * FLOATS_INCREASE_FACTOR
    if num_floats > EXTRA_FLOATS_LIMIT:
        raise TooManyUnprocessedFloatsException(f'tried increasing max floats, '
                                                f'but hit limit of {EXTRA_FLOATS_LIMIT} floats')
    extra_float = ExtraFloats(num_floats)
    return extra_float, num_floats

# def handle_too_many_floats_exception(latex_str: str, callback: Callable, extra_floats_num: Optional[int] = None,
#                                      **callback_kwargs):
#     from pyexlatex.logic.builder import _build
#
#     if extra_floats_num is None:
#         num_floats = DEFAULT_EXTRA_FLOATS
#     else:
#         num_floats = extra_floats_num * FLOATS_INCREASE_FACTOR
#     if num_floats > EXTRA_FLOATS_LIMIT:
#         raise TooManyUnprocessedFloatsException(f'tried increasing max floats, '
#                                                 f'but hit limit of {EXTRA_FLOATS_LIMIT} floats')
#     extra_float = ExtraFloats(num_floats)
#     modified_str = _build([extra_float, latex_str])
#     warnings.warn(f'could not create pdf due to too many unprocessed floats. trying again '
#                   f'with {num_floats} extra floats')
#     return callback(modified_str, extra_floats_num=num_floats, **callback_kwargs)



from typing import Optional, Union, Sequence, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.models.item import Item
from mixins.repr import ReprMixin
from pyexlatex.models.lists.item import ListItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.format.fills import VFill
from pyexlatex.logic.type_check import item_is_in_allowed_type_strs


class VerticalFillMixin:
    vertical_fill = False

    def vertically_space_content(self, items):
        output = []
        for item in items:
            if not can_be_included_directly_in_list(item):
                output.append(ListItem(item))
            else:
                output.append(item)
            if self.vertical_fill:
                output.append(VFill())
        if self.vertical_fill:
            output = output[:-1]  # strip final vertical fill, only fill inbetween items

        return output


class ListBase(TextAreaMixin, VerticalFillMixin, Item, ReprMixin):
    name = 'list'
    repr_cols = ['contents']

    def __init__(self, items: Sequence, overlay: Optional['Overlay'] = None,
                 vertical_fill: bool = False, **kwargs):
        self.items = items
        self.overlay = overlay
        self.vertical_fill = vertical_fill
        self.add_data_from_content(items)
        self.content = self.vertically_space_content(items)
        env_modifier_overlay = f'[{overlay}]' if overlay is not None else None
        super().__init__(self.name, self.content, env_modifiers=env_modifier_overlay, **kwargs)


def can_be_included_directly_in_list(item) -> bool:
    check_funcs = [
        _can_be_included_directly_in_list_but_is_not_list,
        _item_is_list_or_list_base
    ]
    for func in check_funcs:
        if func(item):
            return True

    return False


def _can_be_included_directly_in_list_but_is_not_list(item) -> bool:
    allowed_values = [
        'SetCounter',
        'Raw',
        'TextSize'
    ]
    return item_is_in_allowed_type_strs(item, allowed_values)


def _item_is_list_or_list_base(item) -> bool:
    allowed_values = [
        'ListItem',
        'ListBase'
    ]
    return item_is_in_allowed_type_strs(item, allowed_values)




from pyexlatex.resume.resume import Resume
from pyexlatex.resume.models.employment import Employment
from pyexlatex.resume.templates.education import Education
from pyexlatex.resume.templates.publication import Publication
from pyexlatex.resume.templates.section import SpacedSection
from pyexlatex.resume.templates.award import Award
from pyexlatex.resume.templates.reference import Reference



from pyexlatex import Template
from pyexlatex.presentation import Frame
from pyexlatex.presentation.beamer.control.atbeginsection import AtBeginSection
from pyexlatex.presentation.beamer.toc import TableOfContents


class TableOfContentsAtBeginSection(Template):

    def __init__(self):
        toc = TableOfContents()
        frame = Frame([toc], title='Table of Contents')
        at_begin = AtBeginSection(frame)
        self.contents = at_begin
        super().__init__()



from typing import Sequence, Tuple, Optional, List, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.graphics.shape import Shape


class Rectangle(Shape):
    """
    Draws a rectangle.
    """
    shape_name = 'rectangle'

    def __init__(self, width: int, height: int, contents: Optional[Any] = None,
                 offset: Tuple[int, int] = (0, 0), content_position: str = 'center',
                 content_offset: Optional[float] = None,
                 shape_options: Optional[List[str]] = None,
                 text_options: Optional[List[str]] = None,
                 overlay: Optional['Overlay'] = None, unit: str = 'cm'):
        self.size = (width, height)
        self.offset = offset
        self.unit = unit
        shape_options = self._get_list_copy_from_list_or_none(shape_options)
        shape_options.extend(self.get_size_options())

        super().__init__(
            contents=contents,
            content_position=content_position,
            content_offset=content_offset,
            shape_options=shape_options,
            text_options=text_options,
            overlay=overlay,
            offset=offset
        )

    def get_size_options(self) -> List[str]:
        return [
            f'minimum width={self.size[0]}{self.unit}',
            f'minimum height={self.size[1]}{self.unit}',
        ]


from pyexlatex.models.lists.base import ListBase


class UnorderedList(ListBase):
    """
    Bullet point items.
    """
    name = 'itemize'
    is_ListBase = True  # so that passing within lists will treat as a list




from typing import Any

from pyexlatex.models.item import ItemBase
from pyexlatex.models.section.base import TextAreaMixin


class LayoutBase(TextAreaMixin, ItemBase):
    content: Any = None  # should be overridden in the subclass

    def __init__(self, content, **kwargs):
        super().__init__(None, content, **kwargs)

    def __str__(self):
        from pyexlatex.logic.builder import _build
        if isinstance(self.content, str):
            return self.content
        elif isinstance(self.content, (list, tuple)):
            return _build(self.content)
        else:
            return str(self.content)



from typing import Optional, Dict
from pyexlatex.models.references.bibtex.base import BibTexEntryBase


class BibTexManual(BibTexEntryBase):
    """
    Biblography document which is a manual.
    """
    item_type = 'manual'
    required_attrs = ['title']
    optional_attrs = ['author', 'organization', 'address', 'edition', 'month', 'year', 'note']

    def __init__(self, item_accessor: str, title: str, author: Optional[str] = None, organization: Optional[str] = None,
                 address: Optional[str] = None, edition: Optional[str] = None,
                 month: Optional[str] = None, year: Optional[str] = None,
                 note: Optional[str] = None):
        self.author = author
        self.title = title
        self.organization = organization
        self.address = address
        self.edition = edition
        self.year = year
        self.month = month
        self.note = note

        super().__init__(item_accessor)



from typing import Optional, Callable, Tuple
import warnings
from pyexlatex.logic.pdf.errors.exc import OutputLoopConsecutiveDeadCycles
from pyexlatex.models.control.maxdeadcycles import MaxDeadCycles

EXTRA_CYCLES_LIMIT = 1000
DEFAULT_EXTRA_CYCLES = 200
CYCLES_INCREASE_FACTOR = 2


def get_max_dead_cycles_and_new_num_cycles(cycles_num: Optional[int] = None) -> Tuple[MaxDeadCycles, int]:
    if cycles_num is None:
        cycles = DEFAULT_EXTRA_CYCLES
    else:
        cycles = cycles_num * CYCLES_INCREASE_FACTOR
    if cycles > EXTRA_CYCLES_LIMIT:
        raise OutputLoopConsecutiveDeadCycles(f'tried increasing max consecutive dead cycles, '
                                                f'but hit limit of {EXTRA_CYCLES_LIMIT} cycles')
    max_cycles = MaxDeadCycles(cycles)
    return max_cycles, cycles


def handle_output_loop_cycles_exception(latex_str: str, callback: Callable, cycles_num: Optional[int] = None,
                                        **callback_kwargs):
    from pyexlatex.logic.builder import _build

    if cycles_num is None:
        cycles = DEFAULT_EXTRA_CYCLES
    else:
        cycles = cycles_num * CYCLES_INCREASE_FACTOR
    if cycles > EXTRA_CYCLES_LIMIT:
        raise OutputLoopConsecutiveDeadCycles(f'tried increasing max consecutive dead cycles, '
                                                f'but hit limit of {EXTRA_CYCLES_LIMIT} cycles')
    max_cycles = MaxDeadCycles(cycles)
    modified_str = _build([max_cycles, latex_str])
    warnings.warn(f'could not create pdf due to too many consecutive dead cycles. trying again with {cycles} cycles')
    return callback(modified_str, cycles_num=cycles, **callback_kwargs)



from typing import Optional
import string
import random
from pyexlatex.models.item import MultiOptionSimpleItem


class ColorDefinition(MultiOptionSimpleItem):
    definition_type: Optional[str] = None
    name = 'definecolor'
    equal_attrs = [
        'color_content',
        'color_name'
    ]

    def __init__(self, color_content: str, color_name: Optional[str] = None):
        if color_name is None:
            color_name = random_color_name()
        self.color_name = color_name
        self.color_content = color_content
        self.init_data()
        self.data.packages.append('xcolor')
        super().__init__(self.name, self.color_name, self.definition_type, self.color_content)


def random_color_name(length: int = 20) -> str:
    out_str = ''
    for _ in range(length):
        out_str += random.choice(string.ascii_letters)
    return out_str



from pyexlatex.models.references.bibtex.style.style import Style


JOF_STYLE = r"""
%%%% jof.bst
%%%%
%%%% This file was patched by Ivo Welch on May 29, 1999 in order to
%%%% handle Journal of Finance style conventions for articles, books,
%%%% and incollections.

%%%% Ivo Welch is *NOT* familiar with this language, so most modifications
%%%% were intelligent guesswork.  If you improve it, please notify
%%%% ivo.welch@agsm.ucla.edu and send your better version for public
%%%% distribution.

%%%% Note: JF has no systematic format for a working paper.
%%%% Note: JF has article titles plain and book titles italic.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Econometrica bibliography style
%%% For use with harvard.sty and LaTeX2e (NFSS) or LaTeX 2.09
%%% Also requires em-cite.sty to patch harvard.sty
%%% 24-aug-94
%%% Revised 11-aug-95
%%%
%%% Based on makebst, but with modifications by
%%% kcb@hss.caltech.edu (KC Border)
%%% Please report any problems to me.
%%%
%%% The main changes from the output of makebst are:
%%% 1. Addition of a function to replace repeated authors with a 3 em dash.
%%% 	(Search for compare.names, this function is portable.)
%%% 2. Modification of output.nonnull to allow commas inside quotation marks
%%%		and no comma after year.
%%% 3.Eliminated feature (bug?) of the harvard style of citation.
%%%		It creates a short label for subsequent citations.  This alleged
%%%		feature shortens both Aliprantis, Border, and Luxemburg (1991) and 
%%%		Aliprantis, Brown, and Burkinshaw (1991) to Aliprantis et~al. (1991).
%%%		This creates the need to label these as
%%%		Aliprantis, Border, and Luxemburg (1991a) and
%%%		Aliprantis, Brown, and Burkinshaw (1991b), despite the fact
%%%		that these entries have different authors.  I used my own format 
%%%		for creating the labels, and used the same label for 
%%%		both the short and long forms.  I know that the 
%%%		_Chicago_Manual_of_Style recommends using Aliprantis et~al,
%%		but Econometrica regularly writes out three authors,
%%		and it is rare to cite an article with more than four authors in
%%%		Econometrica. (I couldn't find an example of even four being cited,
%%%		but I didn't spend a lot of time looking.)
%%% 4. Addition of macros to handle editions. (My own idiosycracy.)
%%% 5. A new entry type called issue for special issues of a journal.
%%%		(Another idiosyncracy.)
%%% 6. Many details to conform to Econometrica's quirks.
%%% 	(Search for KCB )
%%% 7. I wrote this for use with LaTeX2e and the NFSS2.  However, I did
%%%		add some macros so it will work with LaTeX 2.09.
%%%		(Search for \ifx\undefined\emph )
%%% 8. 11-august-1995:
%%% 	Stefan Hochguertel  <S.Hochguertel@kub.nl> reported a number of 
%%%	differences from Econometrica.
%%%	Search for KCB(8-11-95) to find the corrections.
%%%	Wrote em-cite.sty to get multiple citations correct.
%%%
%%% The remainder of this file was generated by makebst, except for
%%% modifications marked with KCB.
%% 
%% This is file `em.bst', generated 
%% on <1994/8/17> with the docstrip utility (2.2h).
%% 
%% The original source files were:
%% 
%% genbst.mbs  (with options: `ay,har,nm-rev1,nmft-sc,dt-beg,yr-par,tit-qq,atit-u,thtit-a,volp-com,edby,blk-com,pp,ed,abr,and-com,nfss')
%% ---------------------------------------- 
%% *** Econometrica style *** 
%% 
%% (Here are the specifications of the source file)
%% \def\filename{genbst.mbs}
%% \def\filedate{1994 July 1}
%% \def\fileversion{1.3}
%%   For use with BibTeX version 0.99a or later, LaTeX version 2.09
%%-------------------------------------------------------------------
%% NOTICE:
%% This file may be used for non-profit purposes.
%% It may not be distributed in exchange for money,
%%   other than distribution costs.
%%
%% The author provides it `as is' and does not guarantee it in any way.
%%
%% Copyright (C) 1994 Patrick W. Daly
%% Max-Planck-Institut f\"ur Aeronomie
%% Postfach 20
%% D-37189 Katlenburg-Lindau
%% Germany
%%
%% E-mail:
%% SPAN--     nsp::linmpi::daly    (note nsp also known as ecd1)
%% Internet-- daly@linmpi.dnet.gwdg.de
%%-----------------------------------------------------------
%% \CharacterTable
%%  {Upper-case    \A\B\C\D\E\F\G\H\I\J\K\L\M\N\O\P\Q\R\S\T\U\V\W\X\Y\Z
%%   Lower-case    \a\b\c\d\e\f\g\h\i\j\k\l\m\n\o\p\q\r\s\t\u\v\w\x\y\z
%%   Digits        \0\1\2\3\4\5\6\7\8\9
%%   Exclamation   \!     Double quote  \"     Hash (number) \#
%%   Dollar        \$     Percent       \%     Ampersand     \&
%%   Acute accent  \'     Left paren    \(     Right paren   \)
%%   Asterisk      \*     Plus          \+     Comma         \,
%%   Minus         \-     Point         \.     Solidus       \/
%%   Colon         \:     Semicolon     \;     Less than     \<
%%   Equals        \=     Greater than  \>     Question mark \?
%%   Commercial at \@     Left bracket  \[     Backslash     \\
%%   Right bracket \]     Circumflex    \^     Underscore    \_
%%   Grave accent  \`     Left brace    \{     Vertical bar  \|
%%   Right brace   \}     Tilde         \~}
%%---------------------------------------------------------------------
 % This is an author-year citation style bibliography. As such, it is
 % non-standard LaTeX, and requires a style option to function properly.
 % The appropriate style file is:   harvard.sty
 % The form of the bibitem entries is
 %   \harvarditem[Jones et al.]{Jones, Baker, and Smith}{1990}{key}...
 % There are style changes that may be specified in the text with
 % the command \citationstyle.
 %   \citeasnoun{key} ==>>    Jones et al. (1990)
 %   \cite{key} ==>>          (Jones et al. 1990)
 %   \cite[chap. 2]{key} ==>> (Jones et al. 1990, chap. 2)
 % On the first citation of any work, the full author list is used.
%%---------------------------------------------------------------------

ENTRY
  { address
    author
    booktitle
    chapter
    edition
    editor
    howpublished
    institution
    journal
    key
    month
    note
    number
    organization
    pages
    publisher
    school
    series
    title
    type
    volume
    year
  }
  {}
  { label extra.label sort.label }	%%%KCB: removed  long.label

INTEGERS { output.state before.all mid.sentence after.sentence after.block }

%%%KCB: new output state, no.comma
INTEGERS { no.comma}

FUNCTION {init.state.consts}
{ #0 'before.all :=
  #1 'mid.sentence :=
  #2 'after.sentence :=
  #3 'after.block :=
  #4 'no.comma :=	%%%KCB
}

STRINGS { s t }


%%%KCB: revise to use new output state
FUNCTION {output.nonnull}
{ 's :=
 output.state no.comma = %%%KCB: if in this state, write a space
 {  " " * write$ before.all 'output.state := }	%%%KCB
 {  output.state mid.sentence =		%%%KCB: add brace at beginning
% Ivo Welch: this has been changed from original    { ", " * write$ }
    { " " * write$ }
    { output.state after.block =
        { add.period$ write$
          newline$
          "\newblock " write$
        }
        { output.state before.all =
            'write$
            { add.period$ " " * write$ }
          if$
        }
      if$
  }
 if$
 mid.sentence 'output.state :=
 }	%%%KCB: matching brace
 if$	%%%KCB: test for no.comma state
  s
}

%%%KCB(8-11-95): version to output colon, not comma
FUNCTION {output.nonnull.colon}
{ 's :=
 output.state no.comma = %%%KCB: if in this state, write a space
 {  " " * write$ before.all 'output.state := }	
 {  output.state mid.sentence =		
    { ": " * write$ }
    { output.state after.block =
        { add.period$ write$
          newline$
          "\newblock " write$
        }
        { output.state before.all =
            'write$
            { add.period$ " " * write$ }
          if$
        }
      if$
  }
 if$
 mid.sentence 'output.state :=
 }	%%%KCB: matching brace
 if$	%%%KCB: test for no.comma state
  s
}



FUNCTION {output}
{ duplicate$ empty$
    'pop$
    'output.nonnull
  if$
}

FUNCTION {output.check}
{ 't :=
  duplicate$ empty$
    { pop$ "empty " t * " in " * cite$ * warning$ }
    'output.nonnull
  if$
}

%%% KCB(8-11-95)
FUNCTION {output.check.colon}
{ 't :=
  duplicate$ empty$
    { pop$ "empty " t * " in " * cite$ * warning$ }
    'output.nonnull.colon
  if$
}

FUNCTION {output.bibitem}
{ newline$
  "\harvarditem[" write$
  label write$
  "]{" label * "}{" * write$	%%%KCB: replaced long.label with label
  year duplicate$ empty$
    { pop$ "????" }
    'skip$
  if$
  extra.label * "}{" * write$
  cite$ write$
  "}" write$
  newline$
  ""
  before.all 'output.state :=
}

FUNCTION {fin.entry}
{ add.period$
  write$
  newline$
}

FUNCTION {new.block}
{ output.state before.all =
    'skip$
    { after.block 'output.state := }
  if$
}

FUNCTION {new.sentence}
{ output.state after.block =
    'skip$
    { output.state before.all =
        'skip$
        { after.sentence 'output.state := }
      if$
    }
  if$
}

FUNCTION {not}
{   { #0 }
    { #1 }
  if$
}

FUNCTION {and}
{   'skip$
    { pop$ #0 }
  if$
}

FUNCTION {or}
{   { pop$ #1 }
    'skip$
  if$
}

FUNCTION {new.block.checkb}
{ empty$
  swap$ empty$
  and
    'skip$
    'new.block
  if$
}

FUNCTION {field.or.null}
{ duplicate$ empty$
    { pop$ "" }
    'skip$
  if$
}

FUNCTION {emphasize}
{ duplicate$ empty$
    { pop$ "" }
    { "\emph{" swap$ * "}" * }
  if$
}

INTEGERS { nameptr namesleft numnames }

%%%KCB:  Here is a portable set of routines to check for repeated authors
%%% and replace them with a 3 em dash
%%% According to _Chicago_Manual_of_Style_, a dash is used to 
%%% replace an author's name only if the entire list is the same, so:
%%% Compare names. See if prev.name is identical to CurrList.
%%% If not, then just output CurrList. 
%%% If so, replace with \BySame
%%% I use a conservative matching scheme, 
%%% and check to see if full names are identical, 
%%% even though we only use initials.  
%%% This may mean that you have to correct your .bib database 
%%% to ensure consistency.
%%%
%%% sample usage:
%%% search for {format.author}

INTEGERS {currNum prevNum}

STRINGS {PrevList CurrList prev.name PrevTemp CurrTemp}

%%%KCB: Initialize prev.name ot some nonsense string:

FUNCTION {init.prev.name}
{
"My puppy Harley" 'prev.name :=
}

%%%KCB: Must EXECUTE {init.prev.name} after READ

%%%KCB: <CurrList> compare.names <modified name list>
FUNCTION {compare.names}
{	prev.name 'PrevList :=
	duplicate$ 'prev.name :=	% make current list the new previous list
	'CurrList :=
    PrevList num.names$ 'prevNum :=
    CurrList num.names$ 'currNum :=
%
   prevNum currNum =	%% If prevNum = currNum
  {	"" 'PrevTemp :=		%% Then take this branch
	"" 'CurrTemp :=
    #1 'nameptr :=	%%% start with first name in each list
	{ nameptr prevNum > not }%% while nameptr <= prevNum
%% Format both lists the same way.
%% Use full names: bib file might need to be corrected
	{
	PrevList nameptr "{ff }{vv~}{ll}{ jj}" format.name$ PrevTemp * 'PrevTemp := 
	CurrList nameptr "{ff }{vv~}{ll}{ jj}" format.name$ CurrTemp * 'CurrTemp :=
	nameptr #1 + 'nameptr := }
    while$ 
%% Now compare :
	PrevTemp CurrTemp =	%% If PrevTemp = CurrTemp
%Ivo Welch: Replaced	{ "\BySame{}" }  with { CurrList }.  JF does not like ---- for author repeats
	{ CurrList }
	{ CurrList }
	if$	
	}
	{ CurrList }	%% If not prevNum = currNum
  if$			%% If prevNum = currNum
}

%%%
%%%
%%%KCB:  End of new routine.


FUNCTION {format.names}
{ 's :=
  #1 'nameptr :=
  s num.names$ 'numnames :=
  numnames 'namesleft :=
    { namesleft #0 > }
    { nameptr #1 >
%ivo welch changed. original was
%        { s nameptr "{f.~}{vv~}{ll}{, jj}" format.name$ 't := }
%        { s nameptr "{vv~}{ll}{, jj}{, f.}" format.name$ 't := }
        { s nameptr "{ff~}{vv~}{ll}{, jj}" format.name$ 't := }
        { s nameptr "{vv~}{ll}{, jj}{, ff}" format.name$ 't := }
      if$
      nameptr #1 >
        {
          namesleft #1 >
            { ", " * t * }
            {
              ", " *
              t "others" =
                { " et~al." * }
                { " \textmd{and} " * t * }	%%%KCB(8-11-95): added \sc
              if$
            }
          if$
        }
        't
      if$
      nameptr #1 + 'nameptr :=
      namesleft #1 - 'namesleft :=
    }
  while$
}


FUNCTION {format.names.ed}
{ 's :=
  #1 'nameptr :=
  s num.names$ 'numnames :=
  numnames 'namesleft :=
    { namesleft #0 > }
    { s nameptr
%ivo welch changed. original was
%      "{f.~}{vv~}{ll}{, jj}"
      "{ff~}{vv~}{ll}{, jj}"
      format.name$ 't :=
      nameptr #1 >
        {
          namesleft #1 >
            { ", " * t * }
            {
              ", " *
              t "others" =
                { " et~al." * }
                { " {\small and} " * t * }	%%%KCB: added \small
              if$
            }
          if$
        }
        't
      if$
      nameptr #1 + 'nameptr :=
      namesleft #1 - 'namesleft :=
    }
  while$
}

FUNCTION {format.key}
{ empty$
    { key field.or.null }
    { "" }
  if$
}

FUNCTION {format.authors}
{ author empty$
    { "" }
    { "\textmd{" author compare.names format.names * "}" * }
  if$
}

FUNCTION {format.editors}
{ editor empty$
    { "" }
    { "\textmd{" editor compare.names format.names * " }" *
      editor num.names$ #1 >
        { " (eds.)" * }
        { " (ed.)" * }
      if$
    }
  if$
}

FUNCTION {format.in.editors}
{ editor empty$
    { "" }
    { editor format.names.ed
    }
  if$
}

FUNCTION {format.title}
{ title empty$
    { "" }
    { title
% changed by Ivo Welch: Original was     "``" swap$ * ",'' " *
      "" swap$ * ", " *
    }
no.comma 'output.state :=   %%%KCB:
  if$
}

FUNCTION {n.dashify}
{ 't :=
  ""
    { t empty$ not }
    { t #1 #1 substring$ "-" =
        { t #1 #2 substring$ "--" = not
            { "--" *
              t #2 global.max$ substring$ 't :=
            }
            {   { t #1 #1 substring$ "-" = }
                { "-" *
                  t #2 global.max$ substring$ 't :=
                }
              while$
            }
          if$
        }
        { t #1 #1 substring$ *
          t #2 global.max$ substring$ 't :=
        }
      if$
    }
  while$
}

FUNCTION {word.in}
{ "in " }

FUNCTION {format.date}
{ year duplicate$ empty$
    { "empty year in " cite$ * "; set to ????" * warning$
       pop$ "????" }
    'skip$
  if$
  ", " swap$ * extra.label * ", " *	%%%KCB: added : 
   before.all 'output.state :=		%%%KCB:
}

FUNCTION {format.btitle}
{
  title  emphasize
  no.comma 'output.state :=		%%%KCB: 
}

FUNCTION {tie.or.space.connect}
{ duplicate$ text.length$ #3 <
    { "~" }
    { " " }
  if$
  swap$ * *
}

FUNCTION {either.or.check}
{ empty$
    'pop$
    { "can't use both " swap$ * " fields in " * cite$ * warning$ }
  if$
}

FUNCTION {format.bvolume}
{ volume empty$
    { "" }
    { "vol." volume tie.or.space.connect
      series empty$
        'skip$
        { " of " * series emphasize * }
      if$
      "volume and number" number either.or.check
    }
  if$
}

FUNCTION {format.number.series}
{ volume empty$
    { number empty$
        { series field.or.null }
        { output.state mid.sentence =
            { "no." }
            { "No." }
          if$
          number tie.or.space.connect
          series empty$
            { "there's a number but no series in " cite$ * warning$ }
            { " in " * series * }
          if$
        }
      if$
    }
    { "" }
  if$
}

FUNCTION {format.edition}
{ edition empty$
    { "" }
    { output.state mid.sentence =
        { edition "l" change.case$ " edn." * }
        { edition "t" change.case$ " edn." * }
      if$
    }
  if$
}

INTEGERS { multiresult }

FUNCTION {multi.page.check}
{ 't :=
  #0 'multiresult :=
    { multiresult not
      t empty$ not
      and
    }
    { t #1 #1 substring$
      duplicate$ "-" =
      swap$ duplicate$ "," =
      swap$ "+" =
      or or
        { #1 'multiresult := }
        { t #2 global.max$ substring$ 't := }
      if$
    }
  while$
  multiresult
}

FUNCTION {format.pages}
{ pages empty$
    { "" }
    { pages multi.page.check
        { "pp." pages n.dashify tie.or.space.connect }
        { "p." pages tie.or.space.connect }
      if$
    }
  if$
}

FUNCTION {format.vol.num.pages}
{ volume field.or.null
% ivo welch: numbers are not used in articles for the JF
%  number empty$
%    'skip$
%    { "(" number * ")" * *
%      volume empty$
%        { "there's a number but no volume in " cite$ * warning$ }
%        'skip$
%      if$
%    }
%  if$
  pages empty$
    'skip$
    { duplicate$ empty$
        { pop$ format.pages }
        { ", " * pages n.dashify * }
      if$
    }
  if$
}

FUNCTION {format.chapter.pages}
{ chapter empty$
    'format.pages
    { type empty$
        { "chap." }
        { type "l" change.case$ }
      if$
      chapter tie.or.space.connect
      pages empty$
        'skip$
        { ", " * format.pages * }
      if$
    }
  if$
}

FUNCTION {format.in.ed.booktitle}
{ booktitle empty$
    { "" }
    { editor empty$
        { word.in booktitle emphasize * }
        { word.in booktitle emphasize * ", ed. by " *	%%%KCB: abbrv ed.
          format.in.editors * }
      if$
    }
  if$
}

FUNCTION {format.in.ed.editor}
{ editor empty$
   { word.in booktitle emphasize * }
% Ivo: changed ed: to eds.:  (note: there is no check for the number of editors!)
   { "in " * format.in.editors * ", eds.: " * booktitle emphasize 	%%%KCB: abbrv ed.
          }
   if$
}

FUNCTION {format.thesis.type}
{ type empty$
    'skip$
    { pop$
      type "t" change.case$
    }
  if$
}

FUNCTION {format.tr.number}
{ type empty$
    { "Working paper, " }	%%%KCB: changed default
    'type
  if$
  number empty$
    { "t" change.case$ }
    { number tie.or.space.connect }
  if$
}

FUNCTION {format.article.crossref}
{
  word.in
  "\cite{" * crossref * "}" *
}

FUNCTION {format.book.crossref}
{ volume empty$
    { "empty volume in " cite$ * "'s crossref of " * crossref * warning$
      word.in
    }
    { "vol." volume tie.or.space.connect
      " of " *
    }
  if$
  "\cite{" * crossref * "}" *
}

FUNCTION {format.incoll.inproc.crossref}
{
  word.in
  "\cite{" * crossref * "}" *
}

FUNCTION {article}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
 crossref missing$
    { journal emphasize "journal" output.check
      format.vol.num.pages output
    }
    { format.article.crossref output.nonnull
      format.pages output
    }
  if$
  note output
  fin.entry
}

FUNCTION {book}
{ output.bibitem
  author empty$
    { format.editors "author and editor" output.check
      editor format.key output
    }
    { format.authors output.nonnull
      crossref missing$
        { "author and editor" editor either.or.check }
        'skip$
      if$
    }
  if$
  format.date "year" output.check
  format.btitle "title" output.check 
	mid.sentence 'output.state :=	%%%KCB:
  crossref missing$
    { format.bvolume output
%%%      format.number.series output
      new.sentence
      publisher "publisher" output.check.colon	% KCB(811-95) follow by colon
% ivo welch: added following line for parentheses 
       "(" swap$  * 
      address "address" output.check		% KCB(8-11-95)  city first
% ivo welch: added following line for parentheses 
       ")" * 
    }
    {
      format.book.crossref output.nonnull
    }
  if$
  format.edition output
  note output
  fin.entry
}

FUNCTION {booklet}
{ output.bibitem
  format.authors output
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  howpublished output
  address output
  note output
  fin.entry
}

FUNCTION {inbook}
{ output.bibitem
  author empty$
    { format.editors "author and editor" output.check
      editor format.key output
    }
    { format.authors output.nonnull
      crossref missing$
        { "author and editor" editor either.or.check }
        'skip$
      if$
    }
  if$
  format.date "year" output.check
  format.btitle "title" output.check
  crossref missing$
    { publisher "publisher" output.check.colon	% KCB(811-95) follow by colon
      address "address" output.check		% KCB(8-11-95)  city first
      format.bvolume output	% KCB(811-95): Yes, Econometrica does this!
      format.chapter.pages "chapter and pages" output.check
%%%      format.number.series output
    }
    { format.book.crossref output.nonnull
      format.chapter.pages "chapter and pages" output.check
    }
  if$
  format.edition output
  note output
  fin.entry
}

FUNCTION {incollection}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  crossref missing$
    { format.in.ed.editor "booktitle" output.check
%      new.sentence
      publisher "publisher" output.check % output.check.colon	% KCB(811-95) follow by colon
% ivo welch: added parenthesis before incollection address 
	"(" swap$ * ", " *
      address "address" output.check		% KCB(8-11-95)  city first
% ivo welch: added parenthesis after address; has a bad space in it, but so what.
	")" output
% ivo welch: deleted all the rest
%      format.edition output	% KCB(811-95): Yes, Econometrica does this!
%      format.bvolume output
%      format.number.series output
%      format.chapter.pages output
     }
    { format.chapter.pages output
      format.incoll.inproc.crossref output.nonnull
    }
  if$
  note output
  fin.entry
}

FUNCTION {inproceedings}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  crossref missing$
    { format.in.ed.booktitle "booktitle" output.check
      format.bvolume output
      format.number.series output
      format.pages output
      address output
      new.sentence
      organization output
      publisher output
    }
    { format.incoll.inproc.crossref output.nonnull
      format.pages output
    }
  if$
  note output
  fin.entry
}

FUNCTION {conference} { inproceedings }

%%%KCB: special issue of a journal with a special editor

FUNCTION {issue}
{ output.bibitem
  format.editors output
  editor format.key output
  format.date "year" output.check
  format.btitle "title" output.check
  crossref missing$
    { journal emphasize "journal" output.check
      format.vol.num.pages output
    }
    { format.article.crossref output.nonnull
      format.pages output
    }
  if$
  note output
  fin.entry
}

FUNCTION {manual}
{ output.bibitem
  format.authors output
  author format.key output
  format.date "year" output.check
  format.btitle "title" output.check
  organization output
  address output
  format.edition output
  note output
  fin.entry
}

FUNCTION {mastersthesis}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  "Master's thesis" format.thesis.type output.nonnull
  school "school" output.check
  address output
  note output
  fin.entry
}

FUNCTION {misc}
{ output.bibitem
  format.authors output
  author format.key output
  format.date "year" output.check
  format.title output
  howpublished output
  note output
  fin.entry
}

FUNCTION {phdthesis}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  "Ph.D. thesis" format.thesis.type output.nonnull
  school "school" output.check
  address output
  note output
  fin.entry
}

FUNCTION {proceedings}
{ output.bibitem
  format.editors output
  editor format.key output
  format.date "year" output.check
  format.btitle "title" output.check
  format.bvolume output
  format.number.series output
  address output
  new.sentence
  organization output
  publisher output
  note output
  fin.entry
}

FUNCTION {techreport}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  format.tr.number output.nonnull
  institution "institution" output.check
  address output
  note output
  fin.entry
}

FUNCTION {unpublished}
{ output.bibitem
  format.authors "author" output.check
  author format.key output
  format.date "year" output.check
  format.title "title" output.check
  note "note" output.check
  fin.entry
}

FUNCTION {default.type} { misc }

MACRO {jan} {"January"}

MACRO {feb} {"February"}

MACRO {mar} {"March"}

MACRO {apr} {"April"}

MACRO {may} {"May"}

MACRO {jun} {"June"}

MACRO {jul} {"July"}

MACRO {aug} {"August"}

MACRO {sep} {"September"}

MACRO {oct} {"October"}

MACRO {nov} {"November"}

MACRO {dec} {"December"}

%%%KCB: removed computer science journal macros
%%%KCB: added following macros for editions
MACRO {first} {"1st "}
MACRO {second} {"2d "}
MACRO {third} {"3d "}
MACRO {fourth} {"4th "}
MACRO {fifth} {"5th "}
MACRO {sixth} {"6th"}
MACRO {seventh} {"7th "}
MACRO {eighth} {"8th "}
MACRO {ninth} {"9th "}
MACRO {tenth} {"10th "}

READ

FUNCTION {sortify}
{ purify$
  "l" change.case$
}

INTEGERS { len }

FUNCTION {chop.word}
{ 's :=
  'len :=
  s #1 len substring$ =
    { s len #1 + global.max$ substring$ }
    's
  if$
}


%%%KCB:  Change format of labels:
FUNCTION {format.lab.names}
{ 's :=
  #1 'nameptr :=
  s num.names$ 'numnames :=
  numnames 'namesleft :=
    { namesleft #0 > }
    { s nameptr "{vv~}{ll}" format.name$ 't :=
      nameptr #1 >
	{ namesleft #1 >
	    { ", " * t * }
	    { numnames #2 >
		{ "," * }
		'skip$
	      if$
	      t "others" =
		{ " et~al." * }
		{ " and " * t * }
	      if$
	    }
	  if$
	}
	't
      if$
      nameptr #1 + 'nameptr :=
      namesleft #1 - 'namesleft :=
    }
  while$
}

%%%KCB: removed FUNCTION {format.long.lab.names} 

FUNCTION {author.key.label}
{ author empty$
    { key empty$
        { cite$ #1 #3 substring$ }
        'key
      if$
    }
    { author format.lab.names }
  if$
}

FUNCTION {author.editor.key.label}
{ author empty$
    { editor empty$
        { key empty$
            { cite$ #1 #3 substring$ }
            'key
          if$
        }
        { editor format.lab.names }
      if$
    }
    { author format.lab.names }
  if$  
}

FUNCTION {editor.key.label}
{ editor empty$
    { key empty$
        { cite$ #1 #3 substring$ }
        'key
      if$
    }
    { editor format.lab.names }
  if$
}

%%%KCB: use entire year field as label

FUNCTION {calc.label}
{ type$ "book" =
  type$ "inbook" =
  or
  type$ "issue" =	%%%KCB: need for new entry type
  or				%%%KCB
    'author.editor.key.label
    { type$ "proceedings" =
	'editor.key.label
	'author.key.label
      if$
    }
  if$
  ", "
  *			
  year field.or.null
  *	
  'label :=
}

FUNCTION {calc.short.label}
{ type$ "book" =
  type$ "inbook" =
  or
    'author.editor.key.label
    { type$ "proceedings" =
        'editor.key.label
        'author.key.label
      if$
    }
  if$
  'label :=
}

%%%KCB: removed FUNCTION {calc.long.label}

FUNCTION {sort.format.names}
{ 's :=
  #1 'nameptr :=
  ""
  s num.names$ 'numnames :=
  numnames 'namesleft :=
    { namesleft #0 > }
    { nameptr #1 >
        { "   " * }
        'skip$
      if$
      s nameptr
      "{vv{ } }{ll{ }}{  f{ }}{  jj{ }}"
      format.name$ 't :=
      nameptr numnames = t "others" = and
        { "et al" * }
        { t sortify * }
      if$
      nameptr #1 + 'nameptr :=
      namesleft #1 - 'namesleft :=
    }
  while$
}

FUNCTION {sort.format.title}
{ 't :=
  "A " #2
    "An " #3
      "The " #4 t chop.word
    chop.word
  chop.word
  sortify
  #1 global.max$ substring$
}

FUNCTION {author.sort}
{ author empty$
    { key empty$
        { "to sort, need author or key in " cite$ * warning$
          ""
        }
        { key sortify }
      if$
    }
    { author sort.format.names }
  if$
}

FUNCTION {author.editor.sort}
{ author empty$
    { editor empty$
        { key empty$
            { "to sort, need author, editor, or key in " cite$ * warning$
              ""
            }
            { key sortify }
          if$
        }
        { editor sort.format.names }
      if$
    }
    { author sort.format.names }
  if$
}

FUNCTION {editor.sort}
{ editor empty$
    { key empty$
        { "to sort, need editor or key in " cite$ * warning$
          ""
        }
        { key sortify }
      if$
    }
    { editor sort.format.names }
  if$
}

FUNCTION {presort}
{ calc.label
  label sortify
  "    "
  *
  type$ "book" =
  type$ "inbook" =
  or
  type$ "issue" =	%%%KCB: need for new entry type
  or				%%%KCB
    'author.editor.sort
    { type$ "proceedings" =
        'editor.sort
        'author.sort
      if$
    }
  if$
  #1 entry.max$ substring$ 
  'sort.label :=
  sort.label
  *
  "    "
  *
  title field.or.null
  sort.format.title
  *
  #1 entry.max$ substring$
  'sort.key$ :=
}

ITERATE {presort}

SORT

STRINGS { last.label next.extra }

INTEGERS { last.extra.num }

FUNCTION {initialize.extra.label.stuff}
{ #0 int.to.chr$ 'last.label :=
  "" 'next.extra :=
  #0 'last.extra.num :=
}

FUNCTION {forward.pass}
{ last.label label =
    { last.extra.num #1 + 'last.extra.num :=
      last.extra.num int.to.chr$ 'extra.label :=
    }
    { "a" chr.to.int$ 'last.extra.num :=
      "" 'extra.label :=
      label 'last.label :=
    }
  if$
}

FUNCTION {reverse.pass}
{ next.extra "b" =
    { "a" 'extra.label := }
    'skip$
  if$
  extra.label 'next.extra :=
}

EXECUTE {initialize.extra.label.stuff}

EXECUTE {init.prev.name}	%%%KCB:  needed for replacing repeated authors

ITERATE {forward.pass}

REVERSE {reverse.pass}

FUNCTION {bib.sort.order}
{ sort.label
  "    "
  *
  year field.or.null sortify
  *
  "    "
  *
  title field.or.null
  sort.format.title
  *
  #1 entry.max$ substring$
  'sort.key$ :=
  calc.short.label
%%%KCB: removed   calc.long.label
}

ITERATE {bib.sort.order}

SORT

FUNCTION {begin.bib}
{ preamble$ empty$
    'skip$
    { preamble$ write$ newline$ }
  if$
%%%KCB: add defn of \BySame, can be overridden in document
"\ifx\undefined\BySame" write$ newline$
  "\newcommand{\BySame}{\leavevmode\rule[.5ex]{3em}{.5pt}\ }"
       write$ newline$
  "\fi" write$ newline$
%%% KCB: For those who still cling to LaTeX 2.09
"\ifx\undefined\textmd" write$ newline$
  "\newcommand{\textmd}[1]{{\sc #1}}"
       write$ newline$
  "\fi" write$ newline$
"\ifx\undefined\emph" write$ newline$
  "\newcommand{\emph}[1]{{\em #1\/}}"
       write$ newline$
  "\fi" write$ newline$
%%%
  "\begin{thebibliography}{}" write$ newline$
}

EXECUTE {begin.bib}

EXECUTE {init.state.consts}

ITERATE {call.type$}

FUNCTION {end.bib}
{ newline$
  "\end{thebibliography}" write$ newline$
"Done." top$	%%%KCB: Signal end for NeXTeX's TexView
}

EXECUTE {end.bib}
%% End of customized bst file 
"""

jof_style = Style('jof', JOF_STYLE)


from pyexlatex.models.item import MultiOptionSimpleItem


class SetLength(MultiOptionSimpleItem):
    name = 'setlength'

    def __init__(self, set_for: str, value: str):
        self.set_for = set_for
        self.value = value

        super().__init__(self.name, set_for, value)
